
class 'fTrack'

function fTrack:__init(track_index)
  self._id = self.generate_self_id()
  fTracks[self.id] = self
  self._patterntracks = { }
  self._track_index = track_index
  self._track = renoise.song().tracks[track_index]
  self:init_patterntracks()
end

fTrack.patterntracks = property(
  function(obj) return obj._patterntracks end,
  function(obj, val) obj._patterntracks = val end
  )

fTrack.id = property(
  function(obj) return obj._id end
  )

fTrack.track = property(
  function(obj) return obj._track end
  )

fTrack.track_index = property(
  function(obj)
    for track_index, track in ipairs(rns.tracks) do
      if rawequal(obj.track, track) then
        return track_index
      end
    end
    return nil
  end
  )

fTrack.generate_self_id = function()
  local id
  repeat id = math.floor(math.random()*1000000)
  until not fTracks[id]
  return id
end

function fTrack:refresh_track_index()
  for track_index, track in ipairs(rns.tracks) do
    if rawequal(track, self.track) then
      self.track_index = track_index
      return
    end
  end
  self:destroy()
end

function fTrack:destroy()
  for i, ftrack in ipairs(fTracks) do
    if ftrack.track_index == self.track_index then
      fTracks[i] = nil
    end
  end
end

function fTrack:init_patterntracks()
  for pattern_index, pattern in ipairs(rns.patterns) do
    local new_patterntrack = fPatterntrack(self.id, pattern_index, self.track_index)
    self.patterntracks[new_patterntrack.id] = new_patterntrack
  end
end

function fTrack:refresh_patterntracks()

  local pattern_indexes = { }
  for pattern_index, pattern in ipairs(rns.patterns) do
    pattern_indexes[pattern_index] = pattern
  end

  -- verify existing patterntracks and remove non-existent
  for patterntrack_id, patterntrack in pairs(self.patterntracks) do
    local pattern_index = patterntrack.pattern_index
    if pattern_index then
      pattern_indexes[pattern_index] = nil
    else
      patterntrack:destroy()
    end
  end

  -- create missing patterntracks
  for pattern_index, pattern in pairs(pattern_indexes) do
    local create = fPatterntrack(self.id, pattern_index, self.track_index)
  end

end

class 'fPatterntrack'

function fPatterntrack:__init(fTrack_id, pattern_index, track_index)
  self._fTrack_id = fTrack_id
  self._id = self:generate_id(fTrack_id)
  fTracks[fTrack_id].patterntracks[self.id] = self
  self._patterntrack = rns:pattern(pattern_index):track(track_index)
  self._pattern_index = pattern_index
  self._notifier = function() print("alias changed in pattern " .. self.pattern_index .. ", track " .. fTracks[self.fTrack_id].track_index) end

  local patterntrack = fTracks[self._fTrack_id].patterntracks[self.id]
  rns:pattern(self._pattern_index):track(fTracks[self._fTrack_id].track_index).alias_pattern_index_observable:add_notifier(patterntrack.notifier)
--  self:refresh_notifier()
end

fPatterntrack.id = property(
  function(obj) return obj._id end
  )

fPatterntrack.fTrack_id = property(
  function(obj) return obj._fTrack_id end
  )

fPatterntrack.notifier = property(
  function(obj) return obj._notifier end
  )

fPatterntrack.patterntrack = property(
  function(obj) return obj._patterntrack end,
  function(obj, val) obj._patterntrack = val end
  )

fPatterntrack.pattern_index = property(
  function(obj)
    for pattern_index, pattern in ipairs(rns.patterns) do
      local patterntrack = rns:pattern(pattern_index):track(fTracks[obj.fTrack_id].track_index)
      if rawequal(patterntrack, obj.patterntrack) then
        return pattern_index
      end
    end
    return nil
  end
  )

function fPatterntrack:generate_id(fTrack_id)
  local id
  repeat id = math.floor(math.random()*1000000)
  until not fTracks[fTrack_id].patterntracks[id]
  return id
end

function fPatterntrack:destroy()
  fTracks[self.fTrack_id].patterntracks[self.id] = nil
end
